from tkinter import *
from tkinter import messagebox
from pathlib import Path
from datetime import datetime

import hdl_generation_library
import hdl_generation_architecture
import hdl_generation_module
import main_window

def run_hdl_generation():
    if design_is_not_ready_for_hdl_generation():
        return
    if main_window.language.get()=="VHDL":
        header = "-- Created by HDL-FSM-Editor at " + datetime.today().ctime() + "\n"
    else:
        header = "// Created by HDL-FSM-Editor at " + datetime.today().ctime() + "\n"
    create_hdl(header)

def design_is_not_ready_for_hdl_generation():
    name = main_window.module_name.get()
    if name.isspace() or name=="":
        messagebox.showerror("Error", 'No module-name is specified,\nso no HDL-generation is possible.')
        return True
    path_string = main_window.generate_path_value.get()
    if path_string.isspace() or path_string=="":
        messagebox.showerror("Error", 'No path for the generation of the HDL-files is specified,\nso no HDL-generation is possible.')
        return True
    else:
        path = Path(path_string)
        if not path.exists():
            messagebox.showerror("Error", 'The specified path (' + path_string + ') for the generation of the HDL-files does not exist,\nso HDL-generation is not possible. See "Path for generated HDL" in the control tab.')
            return True
    reset = main_window.reset_signal_name.get()
    if reset.isspace() or reset=="":
        messagebox.showerror("Error", 'No reset signal name is specified,\nso no HDL-generation is possible.')
        return True
    clock = main_window.clock_signal_name.get()
    if clock.isspace() or clock=="":
        messagebox.showerror("Error", 'No clock signal name is specified,\nso no HDL-generation is possible.')
        return True
    return False

def create_hdl(header):
    if main_window.language.get()=="VHDL":
        entity       = create_entity()
        architecture = hdl_generation_architecture.create_architecture()
    else:
        entity       = create_module_ports()
        architecture = hdl_generation_module.create_module()
    if architecture is None:
        return # No further actions required, because when writing to a file, always an architecture must exist.
    else:
        hdl = write_hdl_file(header, entity, architecture)
        copy_hdl_into_generated_hdl_tab(hdl)

def copy_hdl_into_generated_hdl_tab(hdl):
    main_window.hdl_frame_text.config(state=NORMAL)
    main_window.hdl_frame_text.delete("1.0", END)
    main_window.hdl_frame_text.insert("1.0", hdl)
    main_window.hdl_frame_text.update_highlight_tags(10, ["not_read" , "not_written" , "control" , "datatype" , "function" , "comment"])
    main_window.hdl_frame_text.config(state=DISABLED)
    # Bring the notebook tab with the hdl into the foreground:
    notebook_ids = main_window.notebook.tabs()
    for id in notebook_ids:
        if main_window.notebook.tab(id, option="text")=="generated HDL":
            main_window.notebook.select(id)

def create_entity():
    package_statements = hdl_generation_library.get_text_from_text_widget(main_window.interface_package_text)
    generic_declarations = hdl_generation_library.get_text_from_text_widget(main_window.interface_generics_text)
    if generic_declarations!="":
        generic_declarations = "    generic (\n" + hdl_generation_library.indent_text_by_the_given_number_of_tabs(2,generic_declarations) + "    );\n"
    port_declarations = hdl_generation_library.get_text_from_text_widget(main_window.interface_ports_text)
    if port_declarations!="":
        port_declarations    = "    port (\n"    + hdl_generation_library.indent_text_by_the_given_number_of_tabs(2,port_declarations   ) + "    );\n"
    entity  = package_statements
    entity += "\n"
    entity += "entity " + main_window.module_name.get() + " is\n"
    entity += generic_declarations
    entity += port_declarations
    entity += "end entity;\n"
    return entity

def create_module_ports():
    parameters = hdl_generation_library.get_text_from_text_widget(main_window.interface_generics_text)
    if parameters!="":
        parameters = "    #(parameter\n" + hdl_generation_library.indent_text_by_the_given_number_of_tabs(1,parameters) + "    )\n"
    ports = hdl_generation_library.get_text_from_text_widget(main_window.interface_ports_text)
    if ports!="":
        ports      = "    (\n"  + hdl_generation_library.indent_text_by_the_given_number_of_tabs(2,ports     ) + "    );\n"
    module_ports  = "module " + main_window.module_name.get() + "\n"
    module_ports += parameters
    module_ports += ports
    return module_ports

def write_hdl_file(header, entity, architecture):
    if main_window.select_file_number_text.get()==1:
        if main_window.language.get()=="VHDL":
            file_type = ".vhd"
            comment_string = "--"
        elif main_window.language.get()=="Verilog":
            file_type = ".v"
            comment_string = "//"
        else:
            file_type = ".sv"
            comment_string = "//"
        content  = comment_string + " Filename: " + main_window.module_name.get() + file_type + "\n"
        content += header
        content += entity + "\n"
        content += architecture
        fileobject = open(main_window.generate_path_value.get() + "/" + main_window.module_name.get() + file_type,'w')
        fileobject.write(content)
        fileobject.close()
        content_with_numbers = add_line_numbers(content)
    else:
        content1  = "-- Filename: " + main_window.module_name.get() + "_e.vhd" + "\n"
        content1 += header
        content1 += entity
        fileobject_entity = open(main_window.generate_path_value.get() + "/" + main_window.module_name.get() + "_e.vhd",'w')
        fileobject_entity.write(content1)
        fileobject_entity.close()
        content2  = "-- Filename: " + main_window.module_name.get() + "_fsm.vhd" + "\n"
        content2 += header
        content2 += architecture
        fileobject_architecture = open(main_window.generate_path_value.get() + "/" + main_window.module_name.get() + "_fsm.vhd",'w')
        fileobject_architecture.write(content2)
        fileobject_architecture.close()
        content_with_numbers1 = add_line_numbers(content1)
        content_with_numbers2 = add_line_numbers(content2)
        content_with_numbers = content_with_numbers1 + content_with_numbers2
    return content_with_numbers

def add_line_numbers(text):
    text_lines = text.split("\n")
    text_length_as_string = str(len(text_lines))
    number_of_needed_digits_as_string = str(len(text_length_as_string))
    content_with_numbers = ""
    for line_number, line in enumerate(text_lines, start=1):
        content_with_numbers += format(line_number, "0" + number_of_needed_digits_as_string + "d") + ": " + line + "\n"
    return content_with_numbers
