from tkinter import *

import hdl_generation_library
import hdl_generation_architecture_state_sequence
import hdl_generation_architecture_state_actions
import main_window

def create_architecture():
    [state_signal_type_definition,
    number_of_states]               = create_type_definition_for_the_state_signal()
    [reset_condition, reset_action] = hdl_generation_library.create_reset_condition_and_reset_action()
    if reset_condition is None:
        return # No further actions make sense, as always a reset condition must exist.
    package_statements              = hdl_generation_library.get_text_from_text_widget(main_window.internals_package_text)
    signal_declarations             = hdl_generation_library.get_text_from_text_widget(main_window.internals_architecture_text)
    variable_declarations           = hdl_generation_library.get_text_from_text_widget(main_window.internals_process_clocked_text)
    global_actions_before           = hdl_generation_library.create_global_actions_before()
    if global_actions_before!="":
        global_actions_before = "-- Global Actions before:\n" + global_actions_before
    transition_specifications       = hdl_generation_library.extract_transition_specifications_from_the_graph()
    state_sequence                  = hdl_generation_architecture_state_sequence.create_vhdl_for_the_state_sequence(transition_specifications)
    global_actions_after            = hdl_generation_library.create_global_actions_after()
    if global_actions_after!="":
        global_actions_after = "-- Global Actions after:\n" + global_actions_after
    state_actions_process           = hdl_generation_architecture_state_actions.create_state_action_process()
    concurrent_actions              = hdl_generation_library.create_concurrent_actions()
    if concurrent_actions!="":
        concurrent_actions = "-- Global Actions combinatorial:\n" + concurrent_actions
    architecture  = package_statements
    architecture += "\n"
    architecture += "architecture fsm of " + main_window.module_name.get() + " is\n"
    architecture +=      hdl_generation_library.indent_text_by_the_given_number_of_tabs(1, state_signal_type_definition)
    architecture += "    signal state : t_state;\n"
    architecture +=      hdl_generation_library.indent_text_by_the_given_number_of_tabs(1, signal_declarations)
    architecture += "begin\n"
    architecture += "    p_states: process (" + main_window.reset_signal_name.get() + ", " + main_window.clock_signal_name.get() + ")\n"
    architecture +=      hdl_generation_library.indent_text_by_the_given_number_of_tabs(2, variable_declarations)
    architecture += "    begin\n"
    architecture += "        if " + reset_condition + " then\n"
    architecture +=              hdl_generation_library.indent_text_by_the_given_number_of_tabs(3, reset_action)
    architecture += "        elsif rising_edge(" + main_window.clock_signal_name.get() + ") then\n"
    architecture +=              hdl_generation_library.indent_text_by_the_given_number_of_tabs(3, global_actions_before)
    architecture += "            -- State Machine:\n"
    architecture += "            case state is\n"
    architecture +=                  hdl_generation_library.indent_text_by_the_given_number_of_tabs(4, state_sequence)
    architecture += "            end case;\n"
    architecture +=              hdl_generation_library.indent_text_by_the_given_number_of_tabs(3, global_actions_after)
    architecture += "        end if;\n"
    architecture += "    end process;\n"
    architecture +=      hdl_generation_library.indent_text_by_the_given_number_of_tabs(1, state_actions_process)
    architecture +=      hdl_generation_library.indent_text_by_the_given_number_of_tabs(1, concurrent_actions)
    architecture += "end architecture;\n"
    return architecture

def create_type_definition_for_the_state_signal():
    list_of_all_state_names = hdl_generation_library.get_a_list_of_all_state_names()
    number_of_states = len(list_of_all_state_names)
    if list_of_all_state_names!=[]:
        type_definition = "type t_state is ("
        list_of_all_state_names_reduced_by_last_entry = list_of_all_state_names[:-1]
        for state_name in list_of_all_state_names_reduced_by_last_entry:
            type_definition += state_name +","
        type_definition += list_of_all_state_names[-1] + ");\n"
        return [type_definition, number_of_states]
    else:
        return [None, 0]

