from tkinter import *
from tkinter import messagebox

import hdl_generation_library
import hdl_generation_architecture_state_sequence
import hdl_generation_architecture_state_actions
import main_window
import math
import re

def create_module():
    [state_signal_type_definition,
    number_of_states]               = create_signal_declaration_for_the_state_variable()
    [reset_condition, reset_action] = hdl_generation_library.create_reset_condition_and_reset_action()
    if reset_condition==None: return # No further actions make sense, as always a reset condition must exist.
    signal_declarations             = hdl_generation_library.get_text_from_text_widget(main_window.internals_architecture_text)
    variable_declarations           = hdl_generation_library.get_text_from_text_widget(main_window.internals_process_clocked_text)
    if variable_declarations!="":
        indent_inc = 1
        indent_tab = "    "
        always_begin = "    begin\n"
        always_end   = "    end\n"
    else:
        indent_inc = 0
        indent_tab = ""
        always_begin = ""
        always_end   = ""
    global_actions_before           = hdl_generation_library.create_global_actions_before()
    if global_actions_before!="":
        global_actions_before = "// Global Actions before:\n" + global_actions_before
    transition_specifications       = hdl_generation_library.extract_transition_specifications_from_the_graph()
    state_sequence                  = hdl_generation_architecture_state_sequence.create_verilog_for_the_state_sequence(transition_specifications)
    global_actions_after            = hdl_generation_library.create_global_actions_after()
    if global_actions_after!="":
        global_actions_after = "// Global Actions after:\n" + global_actions_after
    state_actions_process           = hdl_generation_architecture_state_actions.create_state_action_process()
    concurrent_actions              = hdl_generation_library.create_concurrent_actions()
    if concurrent_actions!="":
        concurrent_actions = "// Global Actions combinatorial:\n" +  concurrent_actions

    architecture  =          hdl_generation_library.indent_text_by_the_given_number_of_tabs(1, state_signal_type_definition)
    architecture +=          hdl_generation_library.indent_text_by_the_given_number_of_tabs(1, signal_declarations)
    architecture +=          always_begin
    architecture +=              hdl_generation_library.indent_text_by_the_given_number_of_tabs(1 + indent_inc, variable_declarations)
    architecture += indent_tab + "    always @(posedge " + main_window.clock_signal_name.get() + " or " + get_reset_edge(reset_condition) + " " + main_window.reset_signal_name.get() + ") begin: p_states\n"
    architecture += indent_tab + "        if (" + reset_condition + ") begin\n"
    architecture +=                           hdl_generation_library.indent_text_by_the_given_number_of_tabs(3 + indent_inc, reset_action)
    architecture += indent_tab + "        end\n"
    architecture += indent_tab + "        else begin\n"
    architecture +=                           hdl_generation_library.indent_text_by_the_given_number_of_tabs(3 + indent_inc, global_actions_before)
    architecture += indent_tab + "            // State Machine:\n"
    architecture += indent_tab + "            case (state)\n"
    architecture +=                               hdl_generation_library.indent_text_by_the_given_number_of_tabs(4 + indent_inc, state_sequence)
    architecture += indent_tab + "                default:\n"
    architecture += indent_tab + "                    ;\n"
    architecture += indent_tab + "            endcase\n"
    architecture +=                           hdl_generation_library.indent_text_by_the_given_number_of_tabs(3 + indent_inc, global_actions_after)
    architecture += indent_tab + "        end\n"
    architecture += indent_tab + "    end\n"
    architecture +=          always_end
    architecture +=          hdl_generation_library.indent_text_by_the_given_number_of_tabs(1, state_actions_process)
    architecture +=          hdl_generation_library.indent_text_by_the_given_number_of_tabs(1, concurrent_actions)
    architecture += "endmodule\n"
    return architecture

def create_signal_declaration_for_the_state_variable():
    list_of_all_state_names = hdl_generation_library.get_a_list_of_all_state_names()
    if list_of_all_state_names!=[]:
        number_of_states = len(list_of_all_state_names)
        if main_window.language.get()=="Verilog":
            bit_width_number_of_states = math.ceil(math.log(number_of_states,2))
            high_index = bit_width_number_of_states - 1
            signal_declaration = "reg [" + str(high_index) + ":0] state;\n"
            signal_declaration += "localparam\n"
            for i in range(len(list_of_all_state_names)):
                signal_declaration += "    " + list_of_all_state_names[i] + " = " + str(i) + ",\n" 
            signal_declaration_finished = signal_declaration[:-2] + ";\n" # Substitute the last "," by a ";".
        else:
            name_list = ""
            for name in list_of_all_state_names:
                name_list += name + ", "
            signal_declaration_finished = "typedef enum {" + name_list[:-2] + "} t_state;\nt_state state;\n"
        return [signal_declaration_finished, number_of_states]
    else:
        return [None, 0]

def get_reset_edge(reset_condition):
    reset_condition_mod = hdl_generation_library.remove_comments_and_returns(reset_condition)
    reset_condition_mod = re.sub("\\s", "", reset_condition_mod) # remove blanks
    if reset_condition_mod.endswith("1'b0"):
        return "negedge"
    elif reset_condition_mod.endswith("1'b1"):
        return "posedge"
    else:
        messagebox.showerror("Error", "The reset polarity could not be determined from the reset condition: " + reset_condition_mod)
        return "unknown_edge"
